@@ -1,5 +1,5 @@
 """
-" HACK to fore-reload it from vim with :source %
+" HACK to force-reload it from vim with :source %
 pyx import sys; sys.modules.pop("coverage_highlight", None); import coverage_highlight
 finish
 """
@@ -13,6 +13,14 @@ def get_verbosity():
     return int(vim.eval('&verbose'))
 
 
+def get_cmdheight():
+    return int(vim.eval('&cmdheight'))
+
+
+def get_wrapscan():
+    return int(vim.eval('&wrapscan'))
+
+
 def debug(msg):
     if get_verbosity() >= 2:
         print(msg)
@@ -203,41 +211,27 @@ class Signs(object):
         return None
 
     def find_next_range(self, line):
-        # Rewrite using self.iter_ranges() maybe?
-        signs = iter(self)
-        for sign in signs:
-            if sign['lnum'] == line:
-                line += 1
-            if sign['lnum'] > line:
-                break
-        else:
-            return None
-        first = last = sign['lnum']
-        for sign in signs:
-            if sign['lnum'] == last + 1:
-                last += 1
-        return first, last
+        for first, last in self.iter_ranges():
+            if line < first:
+                return (first, last)
+        return None
 
     def find_prev_range(self, line):
-        # Rewrite using self.iter_ranges() maybe?
-        signs = iter(self)
         prev_range = None
-        first = last = None
-        for sign in signs:
-            if last is None:
-                first = last = sign['lnum']
-            elif sign['lnum'] == last + 1:
-                last += 1
-            else:
-                prev_range = (first, last)
-                first = last = sign['lnum']
-            if sign['lnum'] >= line:
-                break
-        else:
-            if last is not None:
+        for first, last in self.iter_ranges():
+            if last < line:
                 prev_range = (first, last)
         return prev_range
 
+    def find_first_range(self):
+        return next(self.iter_ranges(), None)
+
+    def find_last_range(self):
+        last_range = None
+        for first, last in self.iter_ranges():
+            last_range = first, last
+        return last_range
+
 
 def lazyredraw(fn):
     def wrapped(*args, **kw):
@@ -319,7 +313,7 @@ def parse_coverage_output(output, filename):
         print(output)
 
 
-def parse_full_coverage_output(output):
+def parse_full_coverage_output(output, coverage_dir):
     if not output and get_verbosity() >= 1:
         print("Got no output!")
         return
@@ -336,7 +330,7 @@ def parse_full_coverage_output(output):
             print(line)
         if line.startswith('--------'):
             break
-        filename = line.split()[0]
+        filename = os.path.relpath(os.path.join(coverage_dir, line.split()[0]))
         if not os.path.exists(filename):
             # this is unexpected
             if get_verbosity() >= 1:
@@ -387,9 +381,7 @@ def parse_lines(formatted_list, signs):
 
 def program_in_path(program):
     path = os.environ.get("PATH", os.defpath).split(os.pathsep)
-    path = [os.path.join(dir, program) for dir in path]
-    path = [True for file in path if os.path.isfile(file)]
-    return bool(path)
+    return any(os.path.isfile(os.path.join(dir, program)) for dir in path)
 
 
 def find_coverage_script():
@@ -420,7 +412,11 @@ def find_coverage_file_for(filename):
 
 
 def run_coverage_report(coverage_script, coverage_dir, args=[]):
-    print("Running %s report -m %s" % (os.path.relpath(coverage_script), ' '.join(args)))
+    if get_cmdheight() > 1 or get_verbosity() >= 1:
+        print("Running %s report -m %s" % (
+            os.path.relpath(coverage_script), ' '.join(args)))
+        if get_verbosity() >= 2:
+            print("in %s" % coverage_dir)
     if os.path.exists(coverage_script):
         command = [os.path.abspath(coverage_script)]
     else:
@@ -500,7 +496,7 @@ def highlight_all():
         error('Could not find .coverage for %s.' % filename)
         return
     output = run_coverage_report(coverage_script, coverage_dir)
-    parse_full_coverage_output(output)
+    parse_full_coverage_output(output, coverage_dir)
     cursor_moved(force=True)
 
 
@@ -508,44 +504,74 @@ def jump_to_next():
     signs = Signs()
     row, col = vim.current.window.cursor
     next_range = signs.find_next_range(row)
+    wrapped = False
     if next_range is None:
         current_range = signs.find_current_range(row)
-        if current_range is not None:
-            # if we're at the very last range, jump to the last line
-            first, last = current_range
-            if row != last:
-                vim.command("normal! %dG" % last)
+        if get_wrapscan():
+            next_range = signs.find_first_range()
+            if next_range is None:
+                print("No highlighted lines in buffer")
                 return
-        print("No higlighted lines below cursor")
-        return
+            elif next_range == current_range:
+                next_range = None
+            else:
+                wrapped = True
+        if next_range is None:
+            if current_range is not None:
+                # if we're at the very last range, jump to the last line
+                first, last = current_range
+                if row != last:
+                    vim.command("normal! %dG" % last)
+                    return
+            print("No highlighted lines below cursor")
+            return
     first, last = next_range
     # jump to last line so it's visible, then jump back to 1st line
-    # (this does not always work the way I want but eh)
-    vim.command("normal! %dG" % last)
-    vim.command("normal! %dG" % first)
-    print("{}-{}".format(first, last) if first != last else first)
+    jump_to_range(last, first, wrapped)
 
 
 def jump_to_prev():
     signs = Signs()
     row, col = vim.current.window.cursor
     prev_range = signs.find_prev_range(row)
+    wrapped = False
     if prev_range is None:
         current_range = signs.find_current_range(row)
-        if current_range is not None:
-            # if we're at the very first range, jump to the first line
-            first, last = current_range
-            if row != first:
-                vim.command("normal! %dG" % first)
+        if get_wrapscan():
+            prev_range = signs.find_last_range()
+            if prev_range is None:
+                print("No highlighted lines in buffer")
                 return
-        print("No higlighted lines above cursor")
-        return
+            elif prev_range == current_range:
+                prev_range = None
+            else:
+                wrapped = True
+        if prev_range is None:
+            if current_range is not None:
+                # if we're at the very first range, jump to the first line
+                first, last = current_range
+                if row != first:
+                    vim.command("normal! %dG" % first)
+                    return
+            print("No highlighted lines above cursor")
+            return
     first, last = prev_range
+    # jump to first line so it's visible, then jump back to last line
+    jump_to_range(first, last, wrapped)
+
+
+def jump_to_range(first, last, wrapped):
     # jump to first line so it's visible, then jump back to last line
     # (this does not always work the way I want but eh)
     vim.command("normal! %dG" % first)
     vim.command("normal! %dG" % last)
-    print("{}-{}".format(first, last) if first != last else first)
+    # force a redraw now, because if jumping caused scrolling, vim might
+    # shedule a redraw later and clear the message we want to display
+    vim.command("redraw")
+    # sort because sometimes first and last are swapped, in case I want to
+    # position the cursor at the beginning of the range
+    display = "{}-{}".format(*sorted([first, last])) if first != last else first
+    print("{}{}".format(display, " (wrapped)" if wrapped else ""))
 
 
 def cursor_moved(force=False):
